/* PS/2 keyboard routines by Jurre Hanema, version 1.0
 * ps2kbd.c
 *
 * Released under "THE BEER-WARE LICENSE" (Revision 42):
 * <kipmans@gmail.com> wrote this file. As long as you retain this notice you can 
 * do whatever you want with this stuff. If we meet some day, and you think this 
 * stuff is worth it, you can buy me a beer in return.
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "ps2kbd.h"
#include "keycodes.h"

#define LOGIC_XOR(a, b)	(((a) && !(b)) || ((b) && !(a)))

volatile uint8_t	kbd_bit_n = 1;
volatile uint8_t	kbd_n_bits = 0;
volatile uint8_t	kbd_buffer = 0;
volatile uint8_t	kbd_queue[KBD_BUFSIZE + 1];
volatile uint8_t	kbd_queue_idx = 0;
volatile uint16_t	kbd_status = 0;

// Begin actual implementation
void kbd_init()
{
	kbd_queue[KBD_BUFSIZE] = 0;
	
	// Set interrupts
	KBD_SET_INT();
	KBD_EN_INT();
	
	// Enable pullup on clock
	KBD_CLOCK_PORT |= _BV(KBD_CLOCK_BIT);
}


uint8_t kbd_queue_scancode(volatile uint8_t p)
{
	if(kbd_queue_idx < KBD_BUFSIZE)
	{
		kbd_queue[kbd_queue_idx] = p;
		kbd_queue_idx++;
	} else
		return 0;

	return 1;
}


uint8_t kbd_get_scancode()
{
	uint8_t		i;
	uint8_t		tmp;

	if(kbd_queue_idx > 0)
	{
		tmp = kbd_queue[kbd_queue_idx - 1]; 
		
		for(i = 0; i < KBD_BUFSIZE; i++)
			kbd_queue[i] = kbd_queue[i + 1];
		
		kbd_queue_idx--;
		
		return tmp;
	} else
		return 0;
}


void kbd_send(uint8_t data)
{
	while(kbd_status & KBD_SEND) _delay_ms(5);	// This behaviour isn't the most desirable, but it's the easiest and proved to be reliable.
	
	// Initiate request-to-send, the actual sending of the data
	// is handled in the ISR.
	
	KBD_CLOCK_PORT &= ~_BV(KBD_CLOCK_BIT);
	KBD_CLOCK_DDR |= _BV(KBD_CLOCK_BIT);
	_delay_us(120);
	KBD_DATA_DDR |= _BV(KBD_DATA_BIT);
	KBD_CLOCK_DDR &= ~_BV(KBD_CLOCK_BIT);
	KBD_CLOCK_PORT |= _BV(KBD_CLOCK_BIT);
	
	kbd_bit_n = 1;
	kbd_status |= KBD_SEND;
	kbd_n_bits = 0;
	kbd_buffer = data;
}


void kbd_update_leds()
{
	uint8_t	val = 0;

	if(kbd_status & KBD_CAPS) val |= 0x04;
	if(kbd_status & KBD_NUMLOCK) val |= 0x02;
	if(kbd_status & KBD_SCROLL) val |= 0x01;
	
	kbd_send(0xed);
	kbd_send(val);
}

unsigned char decodeScanCode( unsigned char scanCode, uint16_t flags )
{
	unsigned char offset;
	unsigned char result = NOT_USED;
	unsigned char min = 0;
	unsigned char max = keyCount - 1;
	unsigned char mid;
	unsigned char value;

	while ( min <= max )
	{
		mid = ( min + max ) / 2;
		value = pgm_read_byte( &lookUpTable[mid][0] );
		if ( value == scanCode )
		break;
		if ( scanCode > value )
		min = mid + 1;
		else
		max = mid - 1;
	}


	if ( min <= max )
	{
			offset = 1;
		if ( flags & KBD_SHIFT )   // | KBD_CAPS) && !( (flags & KBD_SHIFT) && (flags & KBD_CAPS) ) )
			offset = 2;
		if ( flags & KBD_CAPS )
			if ( flags & KBD_SHIFT )
				offset = 1;
			else
				offset = 2;
			
		if ( flags & KBD_CTRL )
			offset = 3;
			
		if ( flags & KBD_ALT )
			offset = 4;
			
		if ( flags & KBD_ALT_GR )
			offset = 5;

		if (scanCode == 0x69 || scanCode == 0x6b || scanCode == 0x6c || 
			scanCode >= 0x70 && scanCode <= 0x75 || 
			scanCode == 0x7A || scanCode == 0x7d )
		{
			offset = 1;
			if ( !(flags & KBD_EX) )
				if ( flags & KBD_NUMLOCK )
					if ( flags & KBD_SHIFT && !(flags & KBD_CAPS) )
						offset = 1;
					else
						offset = 2;
				else
					if ( flags & KBD_SHIFT && !(flags & KBD_CAPS) )
						offset = 2;
					else
						offset = 1;
		}
		
		
		result = pgm_read_byte(&lookUpTable[mid][offset]);
	}

	return result;
}

// Hauptfunktion zum Einlesen der Scancods mit Fallunterscheidung
unsigned char kbd_getchar()
{
	uint8_t		sc = 0;
	unsigned char	c;
	
	while((sc = kbd_get_scancode()))
	{
		     if(sc == 0xaa)			// acknowledge
			kbd_status |= KBD_BAT_PASSED;
		else if(sc == 0xf0)			// up code (Taste loslassen)
			kbd_status |= KBD_BREAK;
		else if(sc == 0xe0)			// extended code
			kbd_status |= KBD_EX;
		else if(sc == 0xe1)			// extended code 2 (Nur bei Pause-Taste)
			kbd_status |= KBD_EX_2;
		else						// anderer Code
		{
			if(kbd_status & KBD_EX_2)	// extended code 2 vorher gesetzt? (Nur bei Pause-Taste)
			{
				if(kbd_status & KBD_BREAK)
				{
					if(sc == 0x77)
					{
						kbd_status &= ~KBD_EX_2;	// Rcksetzen EX2
						kbd_status &= ~KBD_BREAK;	// Rcksetzen Break
						return PAUSE;
					}
				}
			} // Ende ext. 2
			else if(kbd_status & KBD_EX)	// extended code vorher gesetzt?
			{
				if(kbd_status & KBD_BREAK)	// up code gespeichert
				{
					kbd_status &= ~KBD_EX;		// Rcksetzen EX
					kbd_status &= ~KBD_BREAK;	// Rcksetzen Break
					if(sc == 0x12 || sc == 0x59)// Shift
					{
						kbd_status &= ~KBD_SHIFT;// Rcksetzen SHIFT
					}
					else if(sc == 0x14)			// R ctrl
					{
						kbd_status &= ~KBD_CTRL;//Rcksetzen CTRL
					}						
					else if(sc == 0x11)			// R alt => "Alt Gr"
					{
						kbd_status &= ~KBD_ALT_GR;//Rcksetzen ALT_GR
					}			
				} // Ende up
				else // kein up code  gespeichert
				{					
					if(sc == 0x12 || sc == 0x59)// Shift
					{
						kbd_status &= ~KBD_EX;	// Rcksetzen EX
						kbd_status |= KBD_SHIFT;// Setzen SHIFT
					}						
					else if(sc == 0x14)			// R ctrl
					{
						kbd_status &= ~KBD_EX;	// Rcksetzen EX
						kbd_status |= KBD_CTRL;	// Setzen CTRL
					}						
					else if(sc == 0x11)			// R alt => "Alt Gr"
					{
						kbd_status &= ~KBD_EX;	// Rcksetzen EX
						kbd_status |= KBD_ALT_GR;//Setzen ALT_GR
					}
					else if(sc == 0x4a)			// Num-Block geteilt (/)
					{
						// im Englischen ist hier keine Anpassung ntig
						// im Deutschen leider schon
						kbd_status &= ~KBD_EX;
						return '/';
					}
					else if(c=decodeScanCode(sc,kbd_status)) // Zeichen ermitteln
					{
						kbd_status &= ~KBD_EX;	// Rcksetzen EX
						return c;
					}
					else
					{
						kbd_status &= ~KBD_EX;	// Rcksetzen EX
					}
				} // Ende else (kein up code)			
			} // Ende ext. Code
			else if(kbd_status & KBD_BREAK)
			{
				kbd_status &= ~KBD_BREAK;	// Rcksetzen BREAK
				
				if(sc == 0x12 || sc == 0x59)	// Shift
					kbd_status &= ~KBD_SHIFT;
				else if(sc == 0x14)				// L Ctrl
					kbd_status &= ~KBD_CTRL;
				else if(sc == 0x11)				// L Alt
					kbd_status &= ~KBD_ALT;
				else if(sc == 0x77 || sc == 0x58 || sc == 0x7e)	// Caps lock, num lock or scroll lock
					kbd_status &= ~KBD_LOCKED;
			} // Ende up
			else
			{
				if(sc == 0x12 || sc == 0x59)	// Shift
					kbd_status |= KBD_SHIFT;
				else if(sc == 0x14)				// L ctrl
					kbd_status |= KBD_CTRL;
				else if(sc == 0x11)				// L alt
					kbd_status |= KBD_ALT;
				else if(sc == 0x77 && !(kbd_status & KBD_LOCKED))		// Num lock
				{
					kbd_status = (kbd_status & KBD_NUMLOCK) ? kbd_status & ~KBD_NUMLOCK : kbd_status | KBD_NUMLOCK;
					kbd_status |= KBD_LOCKED;
					kbd_update_leds();
				}
				else if(sc == 0x58 && !(kbd_status & KBD_LOCKED))		// Caps lock
				{
					kbd_status = (kbd_status & KBD_CAPS) ? kbd_status & ~KBD_CAPS : kbd_status | KBD_CAPS;
					kbd_status |= KBD_LOCKED;
					kbd_update_leds();
				}
				else if(sc == 0x7e && !(kbd_status & KBD_LOCKED))		// Scroll lock
				{
					kbd_status = (kbd_status & KBD_SCROLL) ? kbd_status & ~KBD_SCROLL : kbd_status | KBD_SCROLL;
					kbd_status |= KBD_LOCKED;
					kbd_update_leds();
				}
				else
				{
					 c=decodeScanCode(sc,kbd_status);	// Zeichen ermitteln
					 return c;
				}	
			} // Ende else (kein ext. kein break)
		} // Ende else (anderer Code)
	} // Ende while
	
	return 0;
}


uint16_t kbd_get_status()
{
	return kbd_status;
}


ISR(KBD_INT)
{
	if(kbd_status & KBD_SEND)
	{
		// Send data
		
		if(kbd_bit_n == 9)				// Parity bit
		{
			if(kbd_n_bits & 0x01)
				KBD_DATA_DDR |= _BV(KBD_DATA_BIT);
			else
				KBD_DATA_DDR &= ~_BV(KBD_DATA_BIT);
		} else if(kbd_bit_n == 10)			// Stop bit
			KBD_DATA_DDR &= ~_BV(KBD_DATA_BIT);
		else if(kbd_bit_n == 11)			// ACK bit, set by device
		{
			kbd_buffer = 0;
			kbd_bit_n = 0;
			kbd_status &= ~KBD_SEND;
		} else					// Data bits
		{
			if(kbd_buffer & (1 << (kbd_bit_n - 1)))
			{
				KBD_DATA_DDR &= ~_BV(KBD_DATA_BIT);
				kbd_n_bits++;
			} else
				KBD_DATA_DDR |= _BV(KBD_DATA_BIT);
		}
	} else
	{
		// Receive data
		
		if(kbd_bit_n > 0 && kbd_bit_n < 10)		// Ignore start, parity & stop bit
		{
			if(!bit_is_clear(KBD_DATA_PIN, KBD_DATA_BIT))
				kbd_buffer |= (1 << (kbd_bit_n - 2));
		} else if(kbd_bit_n == 11)
		{
			kbd_queue_scancode(kbd_buffer);
			kbd_buffer = 0;
			kbd_bit_n = 0;
		}
	}
	
	kbd_bit_n++;
	
	KBD_SET_INT();
}
